# coding: utf-8

import logging
import numpy as np
from pathlib import Path
from platform import system
from typing import Optional, Tuple, Union

from .meta_actuator import Actuator
from .._global import OptionalModule

try:
  from Phidget22.Net import Net, PhidgetServerType
  from Phidget22.Devices.Stepper import Stepper, StepperControlMode
  from Phidget22.Devices.DigitalInput import DigitalInput
  from Phidget22.PhidgetException import PhidgetException
except (ImportError, ModuleNotFoundError):
  Net = OptionalModule('Phidget22')
  PhidgetServerType = OptionalModule('Phidget22')
  Stepper = OptionalModule('Phidget22')
  StepperControlMode = OptionalModule('Phidget22')
  DigitalInput = OptionalModule('Phidget22')
  PhidgetException = OptionalModule('Phidget22')


class Phidget4AStepper(Actuator):
  """This class can drive Phidget's 4A Stepper module in speed or in position.

  It relies on the :mod:`Phidget22` module to communicate with the motor
  driver. The driver can deliver up to 4A to the motor, and uses 16 microsteps
  by default. Its acquisition rate is set to 10 values per second in this
  class.

  The distance unit is the `mm` and the time unit is the `s`, so speeds are in
  `mm/s` and accelerations in `mm/s²`.

  .. versionadded:: 2.0.3
  """

  def __init__(self,
               steps_per_mm: float,
               current_limit: float,
               max_acceleration: Optional[float] = None,
               remote: bool = False,
               absolute_mode: bool = False,
               reference_pos: float = 0,
               switch_ports: Tuple[int, ...] = tuple(),
               save_last_pos: bool = False,
               save_pos_folder: Optional[Union[str, Path]] = None) -> None:
    """Sets the args and initializes the parent class.

    Args:
      steps_per_mm: The number of steps necessary to move by `1 mm`. This value
        is used to set a conversion factor on the driver, so that it can be
        driven in `mm` and `s` units.
      current_limit: The maximum current the driver is allowed to deliver to
        the motor, in `A` .
      max_acceleration: If given, sets the maximum acceleration the motor is
        allowed to reach in `mm/s²`.
      remote: Set to :obj:`True` to drive the stepper via a network VINT Hub,
        or to :obj:`False` to drive it via a USB VINT Hub.
      absolute_mode: If :obj:`True`, the target position of the motor will be
        calculated from a reference position. If :obj:`False`, the target
        position of the motor will be calculated from its current position.

        .. versionadded:: 2.0.4
      reference_pos: The position considered as the reference position at the
        beginning of the test. Only takes effect if ``absolute_mode`` is
        :obj:`True`.

        .. versionadded:: 2.0.4
      switch_ports: The indexes of the VINT Hub ports where the switches are
        connected.

        .. versionadded:: 2.0.4
      save_last_pos: If :obj:`True`, the last position of the actuator will be
        saved in a .npy file.

        .. versionadded:: 2.0.4
      save_pos_folder: The path to the folder where to save the last position
        of the motor. Only takes effect if ``save_last_pos`` is :obj:`True`.

        .. versionadded:: 2.0.4
    """

    self._motor: Optional[Stepper] = None

    super().__init__()

    self._steps_per_mm = steps_per_mm
    self._current_limit = current_limit
    self._max_acceleration = max_acceleration
    self._remote = remote
    self._switch_ports = switch_ports
    self._switches = list()

    # The following attribute is set to True to automatically check the state
    # of the switches in the open method to keep the motor to move if a switch
    # has been disconnected or hit.
    # Nevertheless, this check can be bypassed if the motor is used outside
    # from a Crappy loop by setting the attribute to False.
    self._check_switch = True

    self._absolute_mode = absolute_mode
    if self._absolute_mode:
      self._ref_pos = reference_pos

    # Determining the path where to save the last position
    # It depends on the current operating system
    self._path: Optional[Path] = None
    if save_last_pos:
      if save_pos_folder is not None:
        self._path = Path(save_pos_folder)
      elif system() in ('Linux', 'Darwin'):
        self._path = Path.home() / '.machine1000N'
      elif system() == 'Windows':
        self._path = Path.home() / 'AppData' / 'Local' / 'machine1000N'
      else:
        self._save_last_pos = False

    # These buffers store the last known position and speed
    self._last_velocity: Optional[float] = None
    self._last_position: Optional[float] = None

  def open(self) -> None:
    """Sets up the connection to the motor driver as well as the various
    callbacks, and waits for the motor driver to attach."""

    # Setting up the motor driver
    self.log(logging.DEBUG, "Enabling server discovery")
    Net.enableServerDiscovery(PhidgetServerType.PHIDGETSERVER_DEVICEREMOTE)
    self._motor = Stepper()

    # Setting up the switches
    for port in self._switch_ports:
      switch = DigitalInput()
      switch.setIsHubPortDevice(True)
      switch.setHubPort(port)
      self._switches.append(switch)

    # Setting the remote or local status
    self._motor.setIsLocal(not self._remote)
    self._motor.setIsRemote(self._remote)
    for switch in self._switches:
      switch.setIsLocal(not self._remote)
      switch.setIsRemote(self._remote)

    # Setting up the callbacks
    self.log(logging.DEBUG, "Setting the callbacks")
    self._motor.setOnAttachHandler(self._on_attach)
    self._motor.setOnErrorHandler(self._on_error)
    self._motor.setOnVelocityChangeHandler(self._on_velocity_change)
    self._motor.setOnPositionChangeHandler(self._on_position_change)
    for switch in self._switches:
      switch.setOnStateChangeHandler(self._on_end)

    # Opening the connection to the motor driver
    try:
      self.log(logging.DEBUG, "Trying to attach the motor")
      self._motor.openWaitForAttachment(10000)
    except PhidgetException:
      raise TimeoutError("Waited too long for the motor to attach !")

    # Opening the connection to the switches
    for switch in self._switches:
      try:
        self.log(logging.DEBUG, "Trying to attach the switch")
        switch.openWaitForAttachment(10000)
      except PhidgetException:
        raise TimeoutError("Waited too long for the switch to attach !")

    # Energizing the motor
    self._motor.setEngaged(True)

    # Check the state of the switches
    if self._check_switch and not all(
      switch.getState() for switch in self._switches):
      raise ValueError(f"A switch is already hit or disconnected !")

  def set_speed(self, speed: float) -> None:
    """Sets the requested speed for the motor.

    Switches to the correct driving mode if needed.

    Args:
      speed: The speed to reach, in `mm/s`.
    """

    # Switching the control mode if needed
    if not self._motor.getControlMode() == StepperControlMode.CONTROL_MODE_RUN:
      self.log(logging.DEBUG, "Setting the control mode to run")
      self._motor.setControlMode(StepperControlMode.CONTROL_MODE_RUN)

    # Setting the desired velocity
    if abs(speed) > self._motor.getMaxVelocityLimit():
      raise ValueError(f"Cannot set a velocity greater than "
                       f"{self._motor.getMaxVelocityLimit()} mm/s !")
    else:
      self._motor.setVelocityLimit(speed)

  def set_position(self,
                   position: float,
                   speed: Optional[float] = None) -> None:
    """Sets the requested position for the motor.

    Switches to the correct driving mode if needed.

    Args:
      position: The position to reach, in `mm`.
      speed: If not :obj:`None`, the speed to use for moving to the desired
        position.
    """

    # Switching the control mode if needed
    if not (self._motor.getControlMode() ==
            StepperControlMode.CONTROL_MODE_STEP):
      self.log(logging.DEBUG, "Setting the control mode to step")
      self._motor.setControlMode(StepperControlMode.CONTROL_MODE_STEP)

    # Setting the desired velocity if required
    if speed is not None:
      if abs(speed) > self._motor.getMaxVelocityLimit():
        raise ValueError(f"Cannot set a velocity greater than "
                         f"{self._motor.getMaxVelocityLimit()} mm/s !")
      else:
        self._motor.setVelocityLimit(abs(speed))

    # Ensuring that the requested position is valid
    min_pos = self._motor.getMinPosition()
    max_pos = self._motor.getMaxPosition()
    if not min_pos <= position <= max_pos:
      raise ValueError(f"The position value must be between {min_pos} and "
                       f"{max_pos}, got {position} !")

    # Setting the position depending on the driving mode
    if not self._absolute_mode:
      self._motor.setTargetPosition(position)
    else:
      self._motor.setTargetPosition(position - self._ref_pos)

  def get_speed(self) -> Optional[float]:
    """Returns the last known speed of the motor."""

    return self._last_velocity

  def get_position(self) -> Optional[float]:
    """Returns the last known position of the motor."""

    if not self._absolute_mode:
      return self._last_position

    # Cannot perform addition if no known last position
    elif self._last_position is not None:
      return self._last_position + self._ref_pos

  def stop(self) -> None:
    """Deenergizes the motor."""

    if self._motor is not None:
      self._motor.setEngaged(False)

  def close(self) -> None:
    """Closes the connection to the motor."""

    if self._motor is not None:
      if self._path is not None:
        self._path.mkdir(parents=False, exist_ok=True)
        np.save(self._path / 'last_pos.npy', self.get_position())
      self._motor.close()

    for switch in self._switches:
      switch.close()

  def _on_attach(self, _: Stepper) -> None:
    """Callback called when the motor driver attaches to the program.

    It sets the current limit, scale factor, data rate and maximum
    acceleration.
    """

    self.log(logging.INFO, "Motor successfully attached")

    # Setting the current limit for the motor
    min_current = self._motor.getMinCurrentLimit()
    max_current = self._motor.getMaxCurrentLimit()
    if not min_current <= self._current_limit <= max_current:
      raise ValueError(f"The current limit should be between {min_current} $"
                       f"and {max_current} A !")
    else:
      self._motor.setCurrentLimit(self._current_limit)

    # Setting the scale factor and the data rate
    self._motor.setRescaleFactor(1 / 16 / self._steps_per_mm)
    self._motor.setDataInterval(100)

    # Setting the maximum acceleration
    if self._max_acceleration is not None:
      min_accel = self._motor.getMinAcceleration()
      max_accel = self._motor.getMaxAcceleration()
      if not min_accel <= self._max_acceleration <= max_accel:
        raise ValueError(f"The maximum acceleration should be between "
                         f"{min_accel} and {max_accel} m/s² !")
      else:
        self._motor.setAcceleration(self._max_acceleration)

  def _on_error(self, _: Stepper, error_code: int, error: str) -> None:
    """Callback called when the motor driver returns an error."""

    raise RuntimeError(f"Got error with error code {error_code}: {error}")

  def _on_velocity_change(self, _: Stepper, velocity: float) -> None:
    """Callback called when the motor velocity changes."""

    self.log(logging.DEBUG, f"Velocity changed to {velocity}")
    self._last_velocity = velocity

  def _on_position_change(self, _: Stepper, position: float) -> None:
    """Callback called when the motor position changes."""

    self.log(logging.DEBUG, f"Position changed to {position}")
    self._last_position = position

  def _on_end(self, _: DigitalInput, state) -> None:
    """Callback when a switch is hit."""

    if not bool(state):
      self.stop()
      raise ValueError(f"A switch has been hit or disconnected !")



# coding: utf-8

from typing import Dict, Any, Optional
import logging

from .meta_modifier import Modifier


class DownSampler(Modifier):
  """Modifier waiting for a given number of data points to be received, then
  returning only the last received point.

  Similar to :class:`~crappy.modifier.Mean`, except it discards the values 
  that are not transmitted instead of averaging them. Useful for reducing
  the amount of data sent to a Block.
  
  .. versionadded:: 2.0.4
  """

  def __init__(self, n_points: int = 10) -> None:
    """Sets the args and initializes the parent class.

    Args:
      n_points: One value will be sent to the downstream Block only once
        every ``n_points`` received values.
    """

    super().__init__()
    self._n_points: int = n_points
    self._count: int = n_points - 1

  def __call__(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """Receives data from the upstream Block, and if the counter matches the 
    threshold, returns the data.

    If the counter doesn't match the threshold, doesn't return anything and 
    increments the counter.
    """

    self.log(logging.DEBUG, f"Received {data}")

    if self._count == self._n_points - 1:
      self._count = 0
      self.log(logging.DEBUG, f"Sending {data}")
      return data
    
    else:
      self._count += 1
      self.log(logging.DEBUG, "Not returning any data")



# coding: utf-8

from time import time
import numpy as np
from typing import Callable, Dict, Optional
import logging

from .meta_block import Block


def plastic(v: float, yield_strain: float = .005, rate: float = .02) -> float:
  """A basic plastic law given as an example."""

  if v > yield_strain:
    return rate * ((((v - yield_strain) / rate) ** 2 + 1) ** .5 - 1)
  return 0


class FakeMachine(Block):
  """This Block emulates the behavior of a tensile test machine.
  
  It can emulate tensile tests, **not compression tests**. By default, it 
  assumes an elasto-plastic behavior of the tested sample. The main mechanical
  parameters of the material are tunable.
  
  This Block is meant to be driven like a :class:`~crappy.blocks.Machine` 
  Block. It receives speed or position commands from upstream Blocks, and 
  modifies the behavior of the emulated machine accordingly. Its outputs are
  however different from the Machine Blocks, as it outputs the current force, 
  position, and strain of the emulated tensile test machine. The labels
  carrying this data are : ``t(s), F(N), x(mm), Exx(%), Eyy(%)``.
  
  This Block was originally designed for proposing examples that do not require
  any hardware to run, but still display the possibilities of Crappy. It can
  also be used to test a script without actually interacting with hardware.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from *Fake_machine* to *FakeMachine*
  """

  def __init__(self,
               rigidity: float = 8.4E6,
               l0: float = 200,
               max_strain: float = 1.51,
               sigma: Optional[Dict[str, float]] = None,
               nu: float = 0.3,
               plastic_law: Callable[[float], float] = plastic,
               max_speed: float = 5,
               mode: str = 'speed',
               cmd_label: str = 'cmd',
               freq: Optional[float] = 100,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      rigidity: The rigidity of the material, in N, so that 
        :math:`force = rigidity * strain`.

        .. versionchanged:: 2.0.0 renamed from *k* to *rigidity*
      l0: The initial length of the fake sample to test, in mm.
      max_strain: The maximum strain the material can withstand before
        breaking.

        .. versionchanged:: 1.5.10 renamed from *maxstrain* to *max_strain*
      mode: Whether the command sent to the fake machine is a speed or a
        position command. Can be ``'speed'`` or ``'position'``.
      plastic_law: A callable taking the maximum reached strain and returning
        the proportion of the current strain caused by plastic deformation.
      sigma: A :obj:`dict` containing for each label the standard deviation for
        adding noise to the signal. Can be given for part or all of the labels.
        The deviation should be given not normalized, in the same unit as the
        label to which it applies.
      nu: Poisson's ratio of the material.
      cmd_label: The label carrying the command of the fake machine.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.

        .. versionadded:: 1.5.10
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.

        .. versionadded:: 1.5.10
        .. versionchanged:: 2.0.0 renamed from *verbose* to *display_freq*
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.

        .. versionadded:: 2.0.0
    """

    super().__init__()
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    # Setting the mechanical parameters of the material
    self._rigidity = rigidity
    self._l0 = l0
    self._max_strain = max_strain / 100
    self._nu = nu
    self._max_speed = max_speed
    self._sigma = {'F(N)': 50, 'x(mm)': 2e-3,
                   'Exx(%)': 1e-3, 'Eyy(%)': 1e-3} if sigma is None else sigma
    self._plastic_law = plastic_law

    self._mode = mode
    self._cmd_label = cmd_label

    # Creating the mechanical variables
    self._current_pos = 0
    self._prev_t = None
    self._prev_broke_t = time()
    self._plastic_elongation = 0
    self._max_recorded_strain = 0

  def begin(self) -> None:
    """Sends a first value that should be 0."""

    self._prev_t = self.t0
    self._send_values()

  def loop(self) -> None:
    """Receives the latest command value, calculates the new speed and position
    from it, checks whether the sample broke and what the plastic elongation
    is, and finally returns the data."""

    # Getting the latest command
    data = self.recv_last_data(fill_missing=True)
    if self._cmd_label not in data:
      return
    else:
      cmd = data[self._cmd_label]
    t = time()
    delta_t = t - self._prev_t
    self._prev_t = t

    # Calculating the speed based on the command and the mode
    if self._mode == 'speed':
      speed = float(np.sign(cmd)) * float(np.min((self._max_speed,
                                                  np.abs(cmd))))
    elif self._mode == 'position':
      speed = float(np.sign(cmd - self._current_pos)) * float(np.min(
          (self._max_speed, np.abs(cmd - self._current_pos) / delta_t)))
    else:
      raise ValueError(f'Invalid mode : {self._mode} !')

    # Updating the current position
    self._current_pos += speed * delta_t

    # If the max strain is reached, consider that the sample broke
    if self._current_pos / self._l0 > self._max_strain:
      if time() - self._prev_broke_t > 1:
        self._prev_broke_t = time()
        self.log(logging.WARNING, "Sample broke !")
      self._rigidity = 0

    # Compute the plastic elongation separately
    if self._current_pos / self._l0 > self._max_recorded_strain:
      self._max_recorded_strain = self._current_pos / self._l0
      self._plastic_elongation = self._plastic_law(
        self._max_recorded_strain) * self._l0

    # Finally, sending the values
    self._send_values()

  def _add_noise(self, to_send: Dict[str, float]) -> Dict[str, float]:
    """Adds noise to the data to be sent, according to the sigma values
    provided by the user. Then returns the noised data."""

    for label, value in to_send.items():
      if label in self._sigma:
        to_send[label] = np.random.normal(value, self._sigma[label])

    return to_send

  def _send_values(self) -> None:
    """Gathers all the information to be sent, adds noise and send it."""

    to_send = {'t(s)': time() - self.t0,
               'F(N)': (self._current_pos -
                        self._plastic_elongation) / self._l0 * self._rigidity,
               'x(mm)': self._current_pos,
               'Exx(%)': self._current_pos * 100 / self._l0,
               'Eyy(%)': -self._nu * self._current_pos * 100 / self._l0}

    self.send(self._add_noise(to_send))



# coding: utf-8

from time import time
from typing import Optional, List
import logging
from math import log2

from .meta_inout import InOut
from .._global import OptionalModule

try:
  from Phidget22.Net import Net, PhidgetServerType
  from Phidget22.Devices.VoltageRatioInput import VoltageRatioInput
  from Phidget22.PhidgetException import PhidgetException
except (ImportError, ModuleNotFoundError):
  Net = OptionalModule('Phidget22')
  PhidgetServerType = OptionalModule('Phidget22')
  VoltageRatioInput = OptionalModule('Phidget22')
  PhidgetException = OptionalModule('Phidget22')


class PhidgetWheatstoneBridge(InOut):
  """This class can read voltage ratio values from a Phidget Wheatstone Bridge.

  It relies on the :mod:`Phidget22` module to communicate with the load cell
  conditioner. It can acquire values up to `50Hz` with possible gain values
  from `1` to `128`.

  .. versionadded:: 2.0.4
  """

  def __init__(self,
               channel: int = 0,
               hardware_gain: int = 1,
               data_rate: float = 50,
               gain: float = 1,
               offset: float = 0,
               remote: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      channel: The index of the channel from which to acquire data, as an
        :obj:`int`. Should be either `0` or `1`.
      hardware_gain: The gain used by the conditioner for data acquisition. The
        higher the gain, the better the resolution but the narrower the range.
        Should be a power of `2` between `1` and `128`.
      data_rate: The number of samples to acquire per second, as an :obj:`int`.
        Should be between `0.017` (one sample per minute) and `50`.
      gain: A gain to apply to the acquired value, following the formula :
        :math:`returned = gain * acquired + offset`
      offset: An offset to apply to the acquired value, following the formula :
        :math:`returned = gain * acquired + offset`
      remote: Set to :obj:`True` to drive the bridge via a network VINT Hub,
        or to :obj:`False` to drive it via a USB VINT Hub.
    """

    self._load_cell: Optional[VoltageRatioInput] = None

    super().__init__()

    if channel not in (0, 1):
      raise ValueError("The channel should be 0 or 1 !")
    self._channel = channel

    if hardware_gain not in (2 ** i for i in range(8)):
      raise ValueError("The hardware gain should be either 1, 2, 4, 8, 16, "
                       "32, 64 or 128 !")
    self._hardware_gain = hardware_gain

    self._data_rate = data_rate
    self._gain = gain
    self._offset = offset
    self._remote = remote

    self._last_ratio: Optional[float] = None

  def open(self) -> None:
    """Sets up the connection to the load cell conditioner as well as the
    various callbacks, and waits for the load cell conditioner to attach."""

    # Setting up the load cell conditioner
    self.log(logging.DEBUG, "Enabling server discovery")
    Net.enableServerDiscovery(PhidgetServerType.PHIDGETSERVER_DEVICEREMOTE)
    self._load_cell = VoltageRatioInput()
    self._load_cell.setChannel(self._channel)

    # Setting the remote or local status
    if self._remote is True:
      self._load_cell.setIsLocal(False)
      self._load_cell.setIsRemote(True)
    else:
      self._load_cell.setIsLocal(True)
      self._load_cell.setIsRemote(False)

    # Setting up the callbacks
    self._load_cell.setOnAttachHandler(self._on_attach)
    self._load_cell.setOnVoltageRatioChangeHandler(self._on_ratio_change)

    # Opening the connection to the load cell conditioner
    try:
      self.log(logging.DEBUG, "Trying to attach the load cell conditioner")
      self._load_cell.openWaitForAttachment(10000)
    except PhidgetException:
      raise TimeoutError("Waited too long for the motor to attach !")

  def get_data(self) -> Optional[List[float]]:
    """Returns the last known voltage ratio value, adjusted with the gain and
    the offset."""

    if self._last_ratio is not None:
      return [time(), self._gain * self._last_ratio + self._offset]

  def close(self) -> None:
    """Closes the connection to the load cell conditioner."""

    if self._load_cell is not None:
      self._load_cell.close()

  def _on_attach(self, _: VoltageRatioInput) -> None:
    """Callback called when the load cell conditioner attaches to the program.

    Sets the data rate and the hardware gain of the conditioner.
    """

    self.log(logging.INFO, "Load cell conditioner successfully attached")

    # Setting the hardware gain
    self._load_cell.setBridgeGain(int(round(log2(self._hardware_gain), 0) + 1))

    # Setting the data rate
    min_rate = self._load_cell.getMinDataRate()
    max_rate = self._load_cell.getMaxDataRate()
    if not min_rate <= self._data_rate <= max_rate:
      raise ValueError(f"The data rate should be between {min_rate} and "
                       f"{max_rate}, got {self._data_rate} !")
    else:
      self._load_cell.setDataRate(self._data_rate)

  def _on_ratio_change(self, _: VoltageRatioInput, ratio: float) -> None:
    """Callback called when the voltage ratio changes."""

    self.log(logging.DEBUG, f"Voltage ratio changed to {ratio}")
    self._last_ratio = ratio

  def _on_error(self,
                _: VoltageRatioInput,
                error_code: int,
                error: str) -> None:
    """Callback called when the load cell conditioner returns an error."""

    raise RuntimeError(f"Got error with error code {error_code}: {error}")



# coding: utf-8

from time import time
from re import findall
from typing import List, Optional
import logging

from .meta_inout import InOut
from .._global import OptionalModule

try:
  from smbus2 import SMBus
except (ModuleNotFoundError, ImportError):
  smbus2 = OptionalModule("smbus2")

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

try:
  import board
except (ModuleNotFoundError, ImportError):
  board = OptionalModule('board', 'Blinka is necessary to use the I2C bus')

try:
  import busio
except (ModuleNotFoundError, ImportError):
  busio = OptionalModule('busio', 'Blinka is necessary to use the I2C bus')

try:
  import adafruit_ads1x15.ads1115 as ads
  from adafruit_ads1x15.analog_in import AnalogIn
except (ModuleNotFoundError, ImportError):
  ADS = OptionalModule('adafruit_ads1x15',
                       'Blinka is necessary to use the I2C bus')
  AnalogIn = OptionalModule('AnalogIn',
                            'Blinka is necessary to use the I2C bus')

# Register and other configuration values:
Ads1115_pointer_conversion = 0x00
Ads1115_pointer_config = 0x01
Ads1115_pointer_lo_thresh = 0x02
Ads1115_pointer_hi_thresh = 0x03

Ads1115_config_mux = {'A0 - A1': 0x0000,
                      'A0 - A3': 0x1000,
                      'A1 - A3': 0x2000,
                      'A2 - A3': 0x3000,
                      'A0': 0x4000,
                      'A1': 0x5000,
                      'A2': 0x6000,
                      'A3': 0x7000}

Ads1115_config_gain = {0.256: 0x0a00,
                       0.512: 0x0800,
                       1.024: 0x0600,
                       2.048: 0x0400,
                       4.096: 0x0200,
                       6.144: 0x0000}

Ads1115_config_dr = {8: 0x0000,
                     16: 0x0020,
                     32: 0x0040,
                     64: 0x0060,
                     128: 0x0080,
                     250: 0x00A0,
                     475: 0x00C0,
                     860: 0x00E0}

Ads1115_blinka_gain = {0.256: 16,
                       0.512: 8,
                       1.024: 4,
                       2.048: 2,
                       4.096: 1,
                       6.144: 2/3}

Ads1115_backends = ['Pi4', 'blinka']


class ADS1115(InOut):
  """A class for controlling Adafruit's ADS1115 16-bits ADC.

  The ADS1115 InOut is meant for reading conversion values from a 16-bits
  ADS1115 ADC, using the I2C protocol. The output is in Volts by default, but a
  ``gain`` and an ``offset`` can be specified.

  Various settings can be adjusted, like the sample rate, the input mode or the
  voltage range.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from *Ads1115* to *ADS1115*
  """

  def __init__(self,
               backend: str,
               device_address: int = 0x48,
               i2c_port: int = 1,
               sample_rate: int = 128,
               v_range: float = 2.048,
               multiplexer: str = 'A1',
               dry_pin: Optional[int] = None,
               gain: float = 1,
               offset: float = 0) -> None:
    """Checks the validity of the arguments.

    Args:
      backend: The backend for communicating with the ADS1115. Should be one
        of:
        ::

          'Pi4', 'blinka'

        The `'Pi4'` backend is optimized but only works on boards supporting
        the :mod:`smbus2` module, like the Raspberry Pis. The `'blinka'`
        backend may be less performant and requires installing
        :mod:`Adafruit-Blinka` and :mod:`adafruit-circuitpython-ads1x15`, but
        these modules are compatible with and maintained on a wide variety of
        boards.
      device_address: The I2C address of the ADS1115. The default address is
        `0x48`, but it is possible to change this setting using the `ADDR` pin.
      i2c_port: The I2C port over which the ADS1115 should communicate. On most
        Raspberry Pi models the default I2C port is `1`.
      sample_rate: The sample rate for data conversion (in SPS). The available
        sample rates are:
        ::

          8, 16, 32, 64, 128, 250, 475, 860

      v_range: The value (in Volts) of the measured signal corresponding to the
        `0x7FFF` output in bits, i.e. that saturates the sensor. A signal of
        ``-v_range`` Volts gives a `0x8000` output in bits. Available
        ``v_range`` values are:
        ::

          0.256, 0.512, 1.024, 2.048, 4.096, 6.144

      multiplexer: Choice of the inputs to consider. Single-input modes
        actually measure `Ax - GND`. The available ``multiplexer`` values are:
        ::

          'A0', 'A1', 'A2', 'A3',
          'A0 - A1',
          'A0 - A3',
          'A1 - A3',
          'A2 - A3'

      dry_pin: Optionally, reads the end of conversion signal from a GPIO
        rather than from an I2C message. Speeds up the reading and decreases
        the traffic on the I2C bus, but requires one extra wire. With the
        backend `'Pi4'`, give the index of the GPIO in BCM convention. This
        feature is not available with the `'blinka'` backend.

        .. versionadded:: 1.5.8
      gain: Allows to tune the output value according to the formula :
        :math:`output = gain * tension + offset`.
      offset: Allows to tune the output value according to the formula :
        :math:`output = gain * tension + offset`.

    Warning:
      AINx voltages should not be higher than `VDD+0.3V` nor lower than
      `GND-0.3V`. Setting high ``v_range`` values does not allow measuring
      voltages higher than `VDD` !!

    .. versionremoved:: 2.0.0 *ft232h_ser_num* argument
    """

    self._bus = None
    self._dry_pin = None

    if not isinstance(backend, str) or backend not in Ads1115_backends:
      raise ValueError("backend should be in {}".format(Ads1115_backends))
    self._backend = backend

    super().__init__()

    if backend == 'Pi4':
      self._bus = SMBus(i2c_port)
    elif backend == 'blinka':
      i2c = busio.I2C(board.SCL, board.SDA)
      self._ads = ads.ADS1115(i2c)

    self._device_address = device_address

    if v_range not in Ads1115_config_gain:
      raise ValueError("v_range should be in {}".format(list(
        Ads1115_config_gain.keys())))
    else:
      self._v_range = v_range

    if sample_rate not in Ads1115_config_dr:
      raise ValueError("sample_rate should be in {}".format(list(
        Ads1115_config_dr.keys())))
    else:
      self._sample_rate = sample_rate

    if multiplexer not in Ads1115_config_mux:
      raise ValueError("multiplexer should be in {}".format(list(
        Ads1115_config_mux.keys())))
    else:
      self._multiplexer = multiplexer

    if dry_pin is not None:
      if backend == 'ft232h' and not isinstance(dry_pin, str):
        raise TypeError('int_pin should be a string when using the ft232h '
                        'backend !')
      elif backend == 'Pi4' and not isinstance(dry_pin, int):
        raise TypeError('int_pin should be an int when using the Pi4 '
                        'backend !')
      elif backend == 'blinka' and dry_pin is not None:
        self.log(logging.WARNING, "Getting the end of conversion information "
                                  "from the DRY pin is not supported by the "
                                  "blinka backend")
    self._dry_pin = dry_pin
    self._chan = None

    self._gain = gain
    self._offset = offset

  def open(self) -> None:
    """Initializes the I2C communication and the device."""

    if self._backend == 'blinka':
      self._ads.gain = Ads1115_blinka_gain[self._v_range]
      self._ads.data_rate = self._sample_rate
      chan_nr = findall(r'\d', self._multiplexer)
      self.log(logging.INFO, "Opening the ADS1115 with backend blinka")
      if len(chan_nr) == 1:
        self._chan = AnalogIn(self._ads,
                              getattr(ads, 'P{}'.format(chan_nr[0])))
      else:
        self._chan = AnalogIn(self._ads,
                              getattr(ads, 'P{}'.format(chan_nr[0])),
                              getattr(ads, 'P{}'.format(chan_nr[1])))

    else:
      if not self._is_connected():
        raise IOError("The ADS1115 is not connected")

      # Setting the configuration register according to the user values
      init_value = Ads1115_config_mux[self._multiplexer]
      init_value |= Ads1115_config_gain[self._v_range]
      init_value |= 0x0100  # Single shot operating mode
      init_value |= Ads1115_config_dr[self._sample_rate]
      self.log(logging.INFO, "Initializing the ADS1115")
      if self._dry_pin is None:
        # Setting the two last bits to 11 to disable the DRY pin
        init_value |= 0x0003

      self._set_register(Ads1115_pointer_config, init_value)

      # Setting the threshold registers to activate the DRY pin output
      if self._dry_pin is not None:
        self._set_register(Ads1115_pointer_lo_thresh, 0x0000)
        self._set_register(Ads1115_pointer_hi_thresh, 0xFFFF)

      if self._backend == 'Pi4' and self._dry_pin is not None:
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self._dry_pin, GPIO.IN)

  def get_data(self) -> List[float]:
    """Reads the registers containing the conversion result.

    The output is in Volts, unless a gain and offset are applied.

    Returns:
      A :obj:`list` containing the timestamp and the voltage value.
    """

    if self._backend == 'blinka':
      out = [time()]
      value = self._chan.voltage

    else:
      # Reading the config register, and setting it so that the ADS1115 starts
      # a conversion
      ms_byte, ls_byte = self._bus.read_i2c_block_data(self._device_address,
                                                       Ads1115_pointer_config,
                                                       2)
      self._set_register(Ads1115_pointer_config,
                         ((ms_byte | 0x80) << 8) | ls_byte)

      # Waiting for the end of the conversion
      t0 = time()
      while not self._data_available():
        if time() - t0 > 0.5:
          raise TimeoutError('Waited too long for data to be ready !')

      # Reading the output of the conversion
      out = [time()]
      ms_byte, ls_byte = self._bus.read_i2c_block_data(
          self._device_address,
          Ads1115_pointer_conversion,
          2)
      self.log(logging.DEBUG, f"Read {ms_byte, ls_byte} from the device "
                              f"address {self._device_address} at "
                              f"register {Ads1115_pointer_conversion}")

      # Converting the output value into Volts
      value_raw = (ms_byte << 8) | ls_byte
      if ms_byte >> 7:
        value_raw -= 2 ** 16
      value = value_raw * self._v_range / 2 ** 15

    out.append(self._offset + self._gain * value)
    return out

  def close(self) -> None:
    """Closes the I2C bus."""

    if self._backend == 'Pi4' and self._bus is not None:
      self.log(logging.INFO, "Closing the ADS1115")
      self._bus.close()

    if self._backend == 'Pi4' and self._dry_pin is not None:
      self.log(logging.INFO, "CCleaning up the GPIOs")
      GPIO.cleanup()

  def _set_register(self, register_address: int, value: int) -> None:
    """Thin wrapper for writing data to the registers."""

    self.log(logging.DEBUG, f"Writing {[(value >> 8) & 0xFF, value & 0xFF]} "
                            f"to the address {self._device_address} in "
                            f"register {register_address}")
    self._bus.write_i2c_block_data(self._device_address, register_address,
                                   [(value >> 8) & 0xFF, value & 0xFF])

  def _is_connected(self) -> bool:
    """Tries reading a byte from the device.

    Returns:
      :obj:`True` if reading was successful, else :obj:`False`.
    """

    try:
      self._bus.read_byte(self._device_address)
      return True
    except IOError:
      return False

  def _data_available(self) -> bool:
    """Returns :obj:`True` if data is available, :obj:`False` otherwise."""

    # EOC signal from the I2C communication
    if self._dry_pin is None:
      return self._bus.read_i2c_block_data(self._device_address,
                                           Ads1115_pointer_config,
                                           1)[0] & 0x80
    # EOC signal from a GPIO
    elif self._backend == 'ft232h':
      return not bool(self._bus.get_gpio(self._dry_pin))
    else:
      return not bool(GPIO.input(self._dry_pin))



# coding: utf-8

from typing import List, Optional, Union
import numpy as np

from ..._global import OptionalModule
from ..camera_config import Box
from .fields import get_res, get_field, allowed_fields

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class DISCorrelTool:
  """This class is the core of the :class:`~crappy.blocks.DISCorrel` Block.

  It receives images from a :class:`~crappy.camera.Camera` object, and performs 
  Dense Inverse Search correlation on each new image to get fields of interest. 
  It relies on DISFlow for the image correlation, handles the projection of the 
  image on the chosen fields, and calculates the residuals.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from *DISCorrel* to *DISCorrelTool*
  """

  def __init__(self,
               box: Box,
               fields: Optional[List[Union[str, np.ndarray]]] = None,
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               init: bool = True,
               iterations: int = 1,
               gradient_iterations: int = 10,
               patch_size: int = 8,
               patch_stride: int = 3) -> None:
    """Sets the parameters of DISFlow.

    Args:
      box: An instance of the
        :class:`~crappy.tool.camera_config.config_tools.Box` object containing
        the coordinates of the patch on which to perform image correlation.

        .. versionadded:: 2.0.0
      fields: The base of fields to use for the projection, given as a
        :obj:`list` of :obj:`str` or :mod:`numpy` arrays (both types can be
        mixed). Strings are for using automatically-generated fields, the
        available ones are :
        ::

          'x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z'

        If users provide their own fields as arrays, they will be used as-is to
        run the correlation. The user-provided fields must be of shape:
        ::

          (patch_height, patch_width, 2)

        .. versionchanged:: 2.0.5 provided fields can now be numpy arrays
      alpha: Weight of the smoothness term in DISFlow, as a :obj:`float`.
      delta: Weight of the color constancy term in DISFlow, as a :obj:`float`.
      gamma: Weight of the gradient constancy term in DISFlow , as a
        :obj:`float`.
      finest_scale: Finest level of the Gaussian pyramid on which the flow
        is computed in DISFlow (`0` means full scale), as an :obj:`int`.
      init: If :obj:`True`, the last field is used to initialize the
        calculation for the next one.
      iterations: Maximum number of gradient descent iterations in the
        patch inverse search stage in DISFlow, as an :obj:`int`.
      gradient_iterations: Maximum number of gradient descent iterations
        in the patch inverse search stage in DISFlow, as an :obj:`int`.

        .. versionchanged:: 1.5.10
           renamed from *gditerations* to *gradient_iterations*
      patch_size: Size of an image patch for matching in DISFlow
        (in pixels).
      patch_stride: Stride between neighbor patches in DISFlow. Must be
        less than patch size.
    """

    if fields is not None:
      # Splitting the given fields into strings and numpy arrays
      auto_fields = [field for field in fields if isinstance(field, str)]
      user_fields = [field for field in fields
                     if isinstance(field, np.ndarray)]

      # Ensuring all the given fields are either strings or numpy arrays
      if len(fields) != len(auto_fields) + len(user_fields):
        raise TypeError('Correlation fields must be either strings or '
                        'numpy arrays !')

      # Ensuring all the string fields are valid ones
      if not all((field in allowed_fields for field in auto_fields)):
        raise ValueError(f"The only allowed values for the fields given as "
                         f"strings are {allowed_fields}")

      self._fields: List[Union[str, np.ndarray]] = fields
    else:
      self._fields: List[Union[str, np.ndarray]] = ["x", "y", "exx", "eyy"]

    self._init = init

    # These attributes will be set later
    self._img0 = None
    self._height, self._width = None, None
    self.box = box
    self._dis_flow = None
    self._base = None
    self._norm2 = None

    # Setting the parameters of Disflow
    self._dis = cv2.DISOpticalFlow_create(cv2.DISOPTICAL_FLOW_PRESET_FAST)
    self._dis.setVariationalRefinementAlpha(alpha)
    self._dis.setVariationalRefinementDelta(delta)
    self._dis.setVariationalRefinementGamma(gamma)
    self._dis.setFinestScale(finest_scale)
    self._dis.setVariationalRefinementIterations(iterations)
    self._dis.setGradientDescentIterations(gradient_iterations)
    self._dis.setPatchSize(patch_size)
    self._dis.setPatchStride(patch_stride)

  def set_img0(self, img0: np.ndarray) -> None:
    """Sets the initial image to use for the correlation.
    
    .. versionadded:: 1.5.10
    """

    self._img0 = img0
    self._height, self._width, *_ = img0.shape
    self._dis_flow = np.zeros((self._height, self._width, 2))

  def set_box(self) -> None:
    """Sets the region of interest to use for the correlation, and initializes
    other attributes.

    .. versionadded:: 1.5.10
    .. versionremoved:: 2.0.0 *box* argument
    """

    # Sets the bounding box
    x_top, x_bottom, y_left, y_right = self.box.sorted()
    box_height = y_right - y_left
    box_width = x_bottom - x_top

    # Creates and populates the base fields to use for correlation
    fields = np.empty((box_height, box_width, 2, len(self._fields)),
                      dtype=np.float32)
    for i, field in enumerate(self._fields):
      if isinstance(field, str):
        fields[:, :, 0, i], fields[:, :, 1, i] = get_field(field, box_height,
                                                           box_width)
      elif isinstance(field, np.ndarray):
        fields[:, :, :, i] = field

    # These attributes will be used later
    self._base = [fields[:, :, :, i] for i in range(fields.shape[3])]
    self._norm2 = [float(np.sum(base_field ** 2)) for base_field in self._base]

  def get_data(self,
               img: np.ndarray,
               residuals: bool = False) -> List[float]:
    """Processes the input image and returns the requested data in a
    :obj:`list`.

    Args:
      img: The new image to process.
      residuals: Whether the residuals should be calculated or not for the
        image, as a :obj:`bool`.

    Returns:
      A :obj:`list` containing the data to calculate, and the residuals at the
      end if requested.

    .. versionadded:: 1.5.10
    """

    # Making sure the reference image and the base fields were set
    if self._img0 is None:
      raise ValueError("The method set_img0 must be called first for setting "
                       "the reference image !")
    elif self._base is None:
      raise ValueError("The method set_box must be called first for setting "
                       "the region of interest !")

    # Updating the optical flow with the latest image
    if self._init:
      self._dis_flow = self._dis.calc(self._img0, img, self._dis_flow)
    else:
      self._dis_flow = self._dis.calc(self._img0, img, None)

    # Getting the values to calculate as floats
    ret = [float(np.sum(vec * self._crop(self._dis_flow))) / n2 for vec, n2 in
           zip(self._base, self._norm2)]

    # Adding the average residual value if requested
    if residuals:
      ret.append(float(np.average(np.abs(get_res(self._img0, img,
                                                 self._dis_flow)))))

    return ret

  def _crop(self, img: np.ndarray) -> np.ndarray:
    """Crops the image to the given region of interest."""

    x_min, x_max, y_min, y_max = self.box.sorted()
    return img[y_min:y_max, x_min:x_max]



# coding: utf-8

from typing import Optional
from tkinter.messagebox import showerror
import tkinter as tk
import logging
from multiprocessing.queues import Queue

from .camera_config_boxes import CameraConfigBoxes
from .config_tools import Box, SpotsBoxes
from ...camera.meta_camera import Camera
from ...camera.meta_camera.camera_setting import CameraScaleSetting
from ..._global import OptionalModule

try:
  from PIL import Image
except (ModuleNotFoundError, ImportError):
  Image = OptionalModule("pillow")


class DICVEConfig(CameraConfigBoxes):
  """Class similar to :class:`~crappy.tool.camera_config.CameraConfig` but also 
  displaying the bounding boxes of the regions defined as patches.
  
  It relies on the :class:`~crappy.tool.camera_config.config_tools.Box` and 
  :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` tools. It is
  meant to be used for configuring the :class:`~crappy.blocks.DICVE` Block.
  
  .. versionadded:: 1.5.10
  .. versionchanged:: 2.0.0 renamed from *DISVE_config* to *DICVEConfig*
  """

  def __init__(self,
               camera: Camera,
               log_queue: Queue,
               log_level: Optional[int],
               max_freq: Optional[float],
               patches: SpotsBoxes) -> None:
    """Sets the patches and initializes the parent class.

    Args:
      camera: The :class:`~crappy.camera.Camera` object in charge of acquiring 
        the images.
      log_queue: A :obj:`multiprocessing.Queue` for sending the log messages to 
        the main :obj:`~logging.Logger`, only used in Windows.

        .. versionadded:: 2.0.0
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.

        .. versionadded:: 2.0.0
      max_freq: The maximum frequency this window is allowed to loop at. It is
        simply the ``freq`` attribute of the :class:`~crappy.blocks.Camera`
        Block.

        .. versionadded:: 2.0.0
      patches: An instance of
        :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` containing
        the patches to follow for image correlation.
    """

    self._patch_size: Optional[CameraScaleSetting] = None

    super().__init__(camera, log_queue, log_level, max_freq)

    # Setting the patches
    self._spots = patches

  def finish(self) -> None:
    """Method called when the user tries to close the configuration window.

    Check that patches were selected on the image. If not, warns the user and
    prevents him from exiting except with CTRL+C. If not already done by the
    user, also saves the initial length between the patches.
    
    .. versionadded:: 2.0.0
    """

    if self._spots.empty():
      self.log(logging.WARNING, "No patches selected ! Not exiting the "
                                "configuration window")
      showerror('Error !',
                message="Please select patches before exiting the config "
                        "window !\nOr hit CTRL+C to exit Crappy")
      return

    self._spots.save_length()
    self.log(logging.INFO,
             f"Successfully saved L0 ! L0 x : {self._spots.x_l0}, "
             f"L0 y : {self._spots.y_l0}")

    super().stop()

  def _add_settings(self) -> None:
    """Same as in the parent class except it also adds a Path size setting to
    the list of possible settings."""

    self._patch_size = CameraScaleSetting("Patch size (px)", 2, 1024,
                                          default=128)
    self._add_slider_setting(self._patch_size)

    super()._add_settings()

  def _update_settings(self) -> None:
    """Same as in the parent class except it also updates the Path size setting
    in addition to all the other settings."""

    if self._patch_size.value != self._patch_size.tk_var.get():
      self._patch_size.value = self._patch_size.tk_var.get()
    self._patch_size.tk_var.set(self._patch_size.value)

    super()._update_settings()

  def _set_bindings(self) -> None:
    """Binds the left mouse button click to drawing the patches on which to
    perform the image correlation."""

    super()._set_bindings()

    self._img_canvas.bind('<ButtonPress-1>', self._start_box)
    self._img_canvas.bind('<B1-Motion>', self._extend_box)
    self._img_canvas.bind('<ButtonRelease-1>', self._stop_box)

  def _extend_box(self, event: tk.Event) -> None:
    """When the user drags the selection box, updating the four patches being
    drawn."""

    super()._extend_box(event)

    if not self._select_box.no_points() and self._patch_size is not None:
      min_x, max_x, min_y, max_y = self._select_box.sorted()
      size = self._patch_size.value
      if max_x - min_x >= 3 * size and max_y - min_y >= 3 * size:
        self._spots.spot_1 = Box(min_x, min_x + size,
                                 (min_y + max_y - size) // 2,
                                 (min_y + max_y + size) // 2)
        self._spots.spot_2 = Box(max_x - size, max_x,
                                 (min_y + max_y - size) // 2,
                                 (min_y + max_y + size) // 2)
        self._spots.spot_3 = Box((min_x + max_x - size) // 2,
                                 (min_x + max_x + size) // 2,
                                 min_y, min_y + size)
        self._spots.spot_4 = Box((min_x + max_x - size) // 2,
                                 (min_x + max_x + size) // 2,
                                 max_y - size, max_y)

  def _stop_box(self, _: tk.Event) -> None:
    """Simply resets the selection box."""

    # This box is not needed anymore
    self._select_box.reset()

  def _draw_overlay(self) -> None:
    """Draws the detected spots to track on top of the last acquired image."""

    self._draw_spots()

  def _handle_box_outside_img(self, box: Box) -> None:
    """If a patch is outside the image, warning the user and resetting the
    patches."""

    self.log(logging.WARNING, f"The patch {box} is outside the image, "
                              f"resetting the patches")
    self._spots.reset()



# coding: utf-8

import tkinter as tk
from tkinter.messagebox import showerror
from typing import Optional
import logging
from multiprocessing.queues import Queue

from .camera_config_boxes import CameraConfigBoxes
from .config_tools import Box
from ...camera.meta_camera import Camera
from ..._global import OptionalModule

try:
  from PIL import Image
except (ModuleNotFoundError, ImportError):
  Image = OptionalModule("pillow")


class DISCorrelConfig(CameraConfigBoxes):
  """Class similar to :class:`~crappy.tool.camera_config.CameraConfig` but also 
  allowing to select the area on which the correlation will be performed.
  
  It relies on the :class:`~crappy.tool.camera_config.config_tools.Box` tool. 
  It is meant to be used for configuring the :class:`~crappy.blocks.DISCorrel` 
  Block.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from *DISConfig* to *DISCorrelConfig*
  """

  def __init__(self,
               camera: Camera,
               log_queue: Queue,
               log_level: Optional[int],
               max_freq: Optional[float],
               patch: Box) -> None:
    """Initializes the parent class and sets the correlation Box.

    Args:
      camera: The :class:`~crappy.camera.Camera` object in charge of acquiring 
        the images.
      log_queue: A :obj:`multiprocessing.Queue` for sending the log messages to 
        the main :obj:`~logging.Logger`, only used in Windows.

        .. versionadded:: 2.0.0
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.

        .. versionadded:: 2.0.0
      max_freq: The maximum frequency this window is allowed to loop at. It is
        simply the ``freq`` attribute of the :class:`~crappy.blocks.Camera`
        Block.

        .. versionadded:: 2.0.0
      patch: The :class:`~crappy.tool.camera_config.config_tools.Box` container
        that will save the information on the patch where to perform image
        correlation.

        .. versionadded:: 2.0.0
    """

    self._correl_box = patch
    self._draw_correl_box = True

    super().__init__(camera, log_queue, log_level, max_freq)

  @property
  def box(self) -> Box:
    """Returns the :class:`~crappy.tool.camera_config.config_tools.Box` object
    containing the region of interest.
    
    .. versionadded:: 1.5.10
    """

    return self._correl_box

  def finish(self) -> None:
    """Method called when the user tries to close the configuration window.

    Checks that a patch was selected on the image. If not, warns the user and
    prevents him from exiting except with CTRL+C.
    
    .. versionadded:: 2.0.0
    """

    if self.box.no_points():
      self.log(logging.WARNING, "No ROI selected ! Not exiting the "
                                "configuration window")
      showerror('Error !',
                message="Please select a ROI before exiting the config "
                        "window !\nOr hit CTRL+C to exit Crappy")
      return

    super().stop()

  def _set_bindings(self) -> None:
    """Binds the left mouse button click for drawing the box on which the
    correlation will be performed."""

    super()._set_bindings()

    self._img_canvas.bind('<ButtonPress-1>', self._start_box)
    self._img_canvas.bind('<B1-Motion>', self._extend_box)
    self._img_canvas.bind('<ButtonRelease-1>', self._stop_box)

  def _start_box(self, event: tk.Event) -> None:
    """Simply saves the position of the user click, and disables the display of
    the current correl box."""

    super()._start_box(event)

    self._draw_correl_box = False

  def _stop_box(self, _: tk.Event) -> None:
    """Makes sure that the selected region is valid, sets it as the new correl
    box, and enables the display of the correl box."""

    self.log(logging.DEBUG, "Ending the selection box")

    # If it's just a regular click with no dragging, do nothing
    if self._img is None or self._select_box.no_points():
      self._select_box.reset()
      self._draw_correl_box = True
      return

    # The sides need to be sorted before slicing numpy array
    y_left, y_right, x_top, x_bottom = self._select_box.sorted()

    # If the box is flat, resetting it
    if y_left == y_right or x_top == x_bottom:
      self._select_box.reset()
      self._draw_correl_box = True
      return

    # The new correl box is just the copy of the select box
    self._correl_box.update(self._select_box)
    self._select_box.reset()

    self._draw_correl_box = True

  def _draw_overlay(self) -> None:
    """Draws the box to use for performing correlation on top of the last
    acquired image.

    Does not draw the correl box is the user is using the selection box.
    """

    if self._draw_correl_box:
      self._draw_box(self._correl_box)
    self._draw_box(self._select_box)

  def _handle_box_outside_img(self, _: Box) -> None:
    """If the correl box is outside the image, it means that the image size has
    been modified. Simply resetting the correl box then."""

    self._correl_box.reset()



# coding: utf-8

import numpy as np
from typing import Optional, Union, Iterable, Dict
from collections import defaultdict
import logging

from .meta_block import Block


class Synchronizer(Block):
  """This Block takes data from upstream Blocks as input and interpolates it to
  output all the labels on the same timestamps as a reference label.

  This Block is very similar to the :class:`~crappy.blocks.Multiplexer` Block,
  but the `Multiplexer` interpolates data in a time base independent of the
  labels whereas this one takes one label as a reference.

  It can take any number of inputs, provided that they all share a common time
  label. It is also possible to choose which labels are considered for
  interpolation and which are dropped. The interpolation is performed using the
  :obj:`numpy.interp` method.

  This Block is useful for synchronizing data acquired from different sensors,
  in the context when one label should be treated as a reference. This is for
  example the case when synchronizing signals with the output of an image
  processing, to be able to compare all the signals in the time base of the
  image acquisition.

  .. versionadded:: 2.0.5
  """

  def __init__(self,
               reference_label: str,
               time_label: str = 't(s)',
               labels_to_sync: Optional[Union[str, Iterable[str]]] = None,
               freq: Optional[float] = 50,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      reference_label: The label whose timestamps will be taken as a time base
        for performing the interpolation.
      time_label: The label carrying the time information. Should be common to
        all the input Blocks.
      labels_to_sync: An iterable (like a :obj:`list` or a :obj:`tuple`)
        containing the labels to interpolate on the reference label's time
        base, except for the time label that is given separately in the
        ``time_label`` argument. The Block also doesn't output anything until
        data has been received on all these labels. If left to :obj:`None`, all
        the received labels are considered. **It is recommended to always set
        this argument !** It is also possible to give this argument as a single
        :obj:`str` (i.e. not in an iterable).
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    """

    super().__init__()
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    # Initializing the attributes
    self._ref_label = reference_label
    self._time_label = time_label
    self._data: Dict[str, np.ndarray] = defaultdict(self._default_array)

    # Forcing the labels_to_sync into a list
    if labels_to_sync is not None and isinstance(labels_to_sync, str):
      self._to_sync = [labels_to_sync]
    elif labels_to_sync is not None:
      self._to_sync = list(labels_to_sync)
    else:
      self._to_sync = None

  def loop(self) -> None:
    """Receives data, interpolates it, and sends it to the downstream
    Blocks."""

    # Receiving all the upcoming data
    data = self.recv_all_data_raw()

    # Iterating over all the links
    for link_data in data:
      # Only data associated with a time label can be synchronized
      if self._time_label not in link_data:
        continue
      # Extracting the time information from the data
      timestamps = link_data.pop(self._time_label)

      # Adding data from each label in the buffer
      for label, values in link_data.items():
        # Only the labels specified in out_labels is considered
        if (self._to_sync is not None and label not in self._to_sync
            and label != self._ref_label):
          continue

        # Adding the received values to the buffered ones
        self._data[label] = np.concatenate((self._data[label],
                                            np.array((timestamps, values))),
                                           axis=1)
        # Sorting the buffered data, if a same label comes from multiple Links
        self._data[label] = self._data[label][
                            :, self._data[label][0].argsort()]

    # Aborting if there's no data to process
    if not self._data:
      self.log(logging.DEBUG, "No data in the buffer to process")
      return

    # Aborting if there's no data for the reference label
    if self._ref_label not in self._data:
      self.log(logging.DEBUG, "No value for the reference label found in "
                              "the buffer")
      return

    # Making sure there's data for all the requested labels
    if (self._to_sync is not None and
        any(label not in self._data for label in self._to_sync)):
      self.log(logging.DEBUG, "Not all the requested labels received yet")
      return

    # There should also be at least two values for each label
    if any(len(self._data[label][0]) < 2 for label in self._data):
      self.log(logging.DEBUG, "Not at least 2 values for each label in buffer")
      return

    # Getting the minimum time for the interpolation (maximin over all labels)
    min_t = max(data[0, 0] for data in self._data.values())

    # Getting the maximum time for the interpolation (minimax over all labels)
    max_t = min(data[0, -1] for data in self._data.values())

    # Checking if there's a valid time range for interpolation
    if max_t < min_t:
      self.log(logging.DEBUG, "Ranges not matching for interpolation")
      return

    # The array containing the timestamps for interpolating
    interp_times = self._data[self._ref_label][0,
      (self._data[self._ref_label][0] >= min_t) &
      (self._data[self._ref_label][0] <= max_t)]

    # Checking if there are values for the target label in the valid time range
    if not np.any(interp_times):
      self.log(logging.DEBUG,
               "No value of the target label found between the minimum and "
               "maximum possible interpolation times")
      return

    to_send = dict()

    # Building the dict of values to send
    for label, values in self._data.items():

      # Keeping the values of the reference label as they are
      if label == self._ref_label:
        to_send[label] = values[1, :]
      # For all the other labels, performing interpolation
      else:
        to_send[label] = list(np.interp(interp_times, values[0], values[1]))

      # Keeping the last data point before max_t to pass this information on
      last = values[:, values[0] <= max_t][:, -1]
      # Removing the used values from the buffer, except the last data point
      self._data[label] = np.column_stack((last, values[:, values[0] > max_t]))

    if to_send:
      # Adding the time values to the dict of values to send
      to_send[self._time_label] = list(interp_times)

      # Sending the values
      for i, _ in enumerate(interp_times):
        self.send({label: values[i] for label, values in to_send.items()})

  @staticmethod
  def _default_array() -> np.ndarray:
    """Helper function for the default dict."""

    return np.array(([], []))
